# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
from functools import wraps
from logging import getLogger

from django.contrib.auth.decorators import user_passes_test
from django.core.exceptions import PermissionDenied
from django.utils.decorators import available_attrs


logger = getLogger('pstore.audit')


def audit_view(description, mutates=False):
    """
    Decorator to log the calling of this view.  Usage::

        @nonce_required
        @require_POST
        @audit_view('bakes pizza', mutates=True)
        def baking_pizza_view(request, user):
            # ...
    """
    def decorator(func):
        @wraps(func, assigned=available_attrs(func))
        def inner(request, *args, **kwargs):
            humanargs = ', '.join(list(args) +
                                  ['%s=%s' % (k, v)
                                   for k, v in kwargs.items()])
            address = str(request.META.get('REMOTE_ADDR', 'UNKNOWN_IP'))
            if request.user.is_anonymous():
                message = (u'Anonymous user on %s %s %s' %
                           (address, description, humanargs))
            else:
                username = request.user.username.replace(' ', '_')
                message = (u'User %s on %s %s %s' %
                           (username, address, description, humanargs))

            if mutates:
                # Use 'warning' as there is no 'notice' level. And encode
                # as UTF-8 because the automatic encoding would add a BOM
                # before the message. See:
                # http://serverfault.com/questions/407643/
                #       rsyslog-update-on-amazon-linux-suddenly-treats-
                #       info-level-messages-as-emerg
                logger.warning(message.encode('utf-8'))
            else:
                logger.info(message.encode('utf-8'))

            return func(request, *args, **kwargs)
        return inner

    return decorator


def is_logged_in_with_nonce(user):
    """
    Raises a 403 error in case the user is not logged in.
    """
    if user.is_anonymous():
        raise PermissionDenied('User is not logged in')
    if not user.used_nonce:
        raise PermissionDenied('User did not log in with a nonce')
    return True
nonce_required = user_passes_test(is_logged_in_with_nonce)



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
from base64 import b64decode, b64encode

from django.conf import settings
from django.db import models


def _db_engines():
    return set(i['ENGINE'].rsplit('.', 1)[-1]
               for i in settings.DATABASES.values())
_is_mysql = all(i == 'mysql' for i in _db_engines())

if not _is_mysql:
    from warnings import warn
    warn('Not using MySQL engine, you get encoded blob performance')


class Model(models.Model):
    """
    Abstract Django Model that adds default created/modified fields and
    a clone method.
    """
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        """Django metaclass information."""
        abstract = True


class AsciiField(models.CharField):
    """
    MySQL ASCII field.

    Please run the Django unit tests to see if it works as advertised.
    """
    description = 'ASCII'

    def get_prep_value(self, value):
        return value.encode('ascii', 'replace')

    if _is_mysql:
        def db_type(self, **kwargs):
            return 'VARBINARY(%s)' % (self.max_length,)
    else:
        def db_type(self, **kwargs):
            return 'VARCHAR(%s)' % (self.max_length,)


class BlobField(models.Field):
    """
    MySQL BLOB field.

    Please run the Django unit tests to see if it works as advertised.
    """
    description = 'Binary'

    # For the MySQL version we don't need this. The to_python and
    # value_to_string methods will get called for serializing and
    # deserializing fixtures only.
    # For the SQLite3 version, we store the values in base64 in the DB,
    # so to_python must always get called.
    if not _is_mysql:
        __metaclass__ = models.SubfieldBase

    def to_python(self, value):
        """
        Differentiates between unicode and binary strings! unicode values are
        expected to be serialized input from fixtures (and thus base64
        encoded). Binary string values are left as-is.
        """
        if isinstance(value, unicode):
            return b64decode(value)
        return value

    def value_to_string(self, obj):
        """
        Return unicode to flag that we're dealing with serialized data. This
        can be used in fixtures.
        """
        value = self._get_val_from_obj(obj)
        return unicode(b64encode(value))

    if _is_mysql:
        def db_type(self, **kwargs):
            # We use a LONGBLOB which can hold up to 4GB of bytes. A
            # MEDIUMBLOB of max 16MB should probably be enough, but we
            # don't want to add an arbitrary limit there.
            return 'LONGBLOB'
    else:
        def get_prep_value(self, value):
            if value is None:
                return None
            return b64encode(value)

        def db_type(self, **kwargs):
            return 'BLOB'



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
from django.conf import settings
from django.http import HttpResponse
from django.utils.encoding import force_unicode
from django.utils.functional import Promise
from django.utils.simplejson import JSONEncoder, dumps


class _JsonEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Promise):
            return force_unicode(obj)
        return obj


class JsonResponse(HttpResponse):
    """
    Output response as json.
    """
    def __init__(self, request, json_response):
        if not isinstance(json_response, basestring):
            json_kwargs = {
                'check_circular': False,
                'cls': _JsonEncoder,
                'ensure_ascii': False,
                'indent': 2,
                'sort_keys': True,
            }
            if not settings.DEBUG:
                json_kwargs.update({'indent': None, 'separators': (',', ':')})
            json_response = dumps(json_response, **json_kwargs)
        super(JsonResponse, self).__init__(json_response,
                                           content_type='application/json')



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
from django.contrib import admin
from django.views.generic.base import RedirectView

try:
    # Django 1.4+
    from django.conf.urls import include, patterns
except ImportError:
    # Django 1.3-
    from django.conf.urls.defaults import include, patterns


admin.autodiscover()


urlpatterns = patterns('pstore',  # noqa
    # Object identifiers and property names may need to be escaped to allow for
    # the slash (/).

    ###########################################################################
    # JSON interface, for smaller items of data.
    ###########################################################################

    # Get verbose info about a single object.
    (r'^object/(?P<object_identifier>[^/]+).js$',
     'views_js.get_object'),
    # Get a (partial) listing of objects.
    (r'^objects.js$',
     'views_js.list_objects'),
    # Get a (partial) listing of users.
    (r'^users.js$',
     'views_js.list_users'),

    # Get DB consistency report.
    (r'^validate.js$',
     'views_js.validate'),

    ###########################################################################
    # Binary interface, for properties (which can be large)
    ###########################################################################

    # Get a new nonce.
    (r'^nonce.bin$',
     'views_bin.create_nonce'),

    # Get a single object property.
    (r'^propget/(?P<object_identifier>[^/]+)/(?P<property_name>[^/]+).bin$',
     'views_bin.get_property'),
    # Create object and/or set property.
    (r'^propset/(?P<object_identifier>[^/]+)/(?P<property_name>[^/]+).bin$',
     'views_bin.set_property'),

    # Update all shared (encrypted) properties at once. Used when
    # adding/revoking user permissions. (TODO: create a separate one, usable
    # for revoking only? OR: allow update to specify the "bogus" file so that
    # it keeps the original for that user.)
    (r'^propupd/(?P<object_identifier>[^/]+).bin$',
     'views_bin.update_properties'),

    # Search for properties.
    (r'^propsearch.js$',
     'views_js.search_properties'),

    ###########################################################################
    # Admin interface
    ###########################################################################

    # Use / as the admin path (only if this is the only app in the project)
    # (point people to the right url.. fails to work if STATIC_URL is '/')
    (r'^admin(/.*)$', RedirectView.as_view(url='/', permanent=False)),
    (r'', include(admin.site.urls)),
)



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
from datetime import datetime, timedelta
from os import chmod, unlink

from django.contrib.auth.models import User
from django.core.exceptions import PermissionDenied
from django.db import connection
from django.db.models import Q
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.views.decorators.http import require_GET, require_POST

from pstorelib.bytes import BytesIO
from pstorelib.server import urlunquote

from pstore.decorators import audit_view, nonce_required
from pstore.http import EncryptedResponse, HttpError, VoidResponse
from pstore.models import Nonce, Object, ObjectPerm, Property
from pstore.security import get_object_or_403


# NOTE: If the user is a superuser, the results to certain lookups will be 200
# or 404. If the user is a mere mortal, lookups that do not return 200 will
# return 403.


def create_property(object, property, file, user):
    # If isinstance(file, InMemoryUploadedFile) then file.field_name contains
    # property as well. The TemporaryUploadedFile doesn't have it though.

    if user:
        ptype = Property.TYPE_SHARED
        if file.size == 0:
            raise HttpError(400, 'encrypted properties cannot be 0 bytes in '
                                 'length')
    else:
        ptype = Property.TYPE_PUBLIC

    # Depending on the file size, we either get an InMemoryUploadedFile or a
    # TemporaryUploadedFile.
    if hasattr(file, 'temporary_file_path'):
        # Tricks! We'll use a custom INSERT on localhost. Otherwise we'll
        # run into the MySQL max_packet_size limit.
        tempname = file.temporary_file_path()
        data = '...loading...'
    else:
        tempname = None
        data = file.read()
        if len(data) != file.size:
            raise HttpError(
                500, 'file read returned wrong amount of bytes',
                ('We expected to read %d bytes at once from %s. We only got '
                 '%d bytes.' % (file.size, file.name, len(data))))

    prop = Property.objects.create(object=object, name=property,
                                   type=ptype, value=data, user=user)

    # If tempname then we've only written a dummy value thusfar. Update it
    # to the real value.
    if tempname:
        # We use LOAD_FILE() because it speeds up the writing by a factor
        # three. We expect you to run the MySQLd on localhost for now..
        if (connection.settings_dict['HOST'] not in
                ('', 'localhost', '127.0.0.1')):
            raise HttpError(
                413, 'request too large (mysqld infrastructure)',
                ('mysqld can only use LOAD_FILE() on localhost and the DB '
                 'server seems to be running on %s' %
                 (connection.settings_dict['HOST'],)))
        # MySQLd must get read powers.
        try:
            chmod(tempname, 0604)
        except Exception, e:
            raise HttpError(
                413, 'request too large (webserver permissions)',
                ('For mysqld to do load LOAD_FILE() on %s, we need to alter '
                 'file permissions, we got: %s' % (tempname, e)))
        # Try to read the file.
        try:
            cursor = connection.cursor()
            cursor.execute('''
                UPDATE pstore_property SET value = LOAD_FILE(%s)
                WHERE id = %s;
            ''', (tempname, prop.id))
        except Exception, e:
            # Tip #1:
            #   /etc/apparmor.d/local/usr.sbin.mysqld:
            #     /tmp/* r,
            #
            # Tip #2:
            #   mysql> UPDATE mysql.user SET file_priv = 'Y'
            #          WHERE user = 'pstore' AND host = 'localhost'
            #          AND file_priv = 'N'; FLUSH PRIVILEGES;
            #
            # Tip #3:
            #   /etc/mysql/my.cnf:
            #     # If you're doing replication, you must use MIXED or
            #     # ROW based replication. Otherwise LOAD_FILE will fail.
            #     binlog_format = MIXED
            #
            raise HttpError(
                413, 'request too large (mysqld permissions)',
                ('mysqld LOAD_FILE failed for %s, check apparmor. Check '
                 'File_Priv mysql permissions, check @@max_allowed_packet, '
                 'check @@secure_file_priv: %s' % (tempname, e)))
        finally:
            # Remove access to the file asap.
            unlink(tempname)
            # Close the cursor here. The http middleware will commit the
            # transaction when/if done.
            cursor.close()


# NOT nonce_required.. obviously..
@require_POST
@audit_view('creates nonce')
def create_nonce(request):
    """
    One of the few views that does not require a valid nonce.
    """
    # Query strings:
    u = request.GET.get('u', None)  # create nonce for a user
    if not u:
        raise Http404('Bad call')

    # FIXME: instead of returning 403, we should hand out a bogus nonce instead
    # >:-) But then we should do that below too..
    # and what to use for publickey.key_type?
    user = get_object_or_403(User, username=u)

    # Prune old nonces.. might as well do that here.
    old = datetime.now() - timedelta(seconds=Nonce.MAX_AGE)
    Nonce.objects.filter(created__lt=old).delete()

    # Check whether this user has created enough nonces already.
    max_queued_nonces = 10
    if Nonce.objects.filter(user=user).count() > max_queued_nonces:
        raise PermissionDenied()

    # Ok, create a new one.
    nonce = Nonce.objects.create(user=user)

    # Response:
    return EncryptedResponse(data=nonce.encrypted,
                             enctype=user.publickey.key_type())


@nonce_required
@require_GET
@audit_view('reads single property')
def get_property(request, object_identifier, property_name):
    # Query strings:
    u = request.GET.get('u', None)  # filter by user

    # Decode object_identifier and property_name:
    object_identifier = urlunquote(object_identifier)
    property_name = urlunquote(property_name)

    # Query:
    qs = Property.objects.filter(
        object__identifier=object_identifier,
        name=property_name)
    if u:
        qs = qs.filter(Q(user=None) | Q(user__username=u))
    else:
        qs = qs.filter(user=None)

    # Check authorization:
    if request.user.has_perm('pstore.view_any_object'):
        pass
    elif not ObjectPerm.objects.filter(
            object__identifier=object_identifier,
            user=request.user).exists():
        raise PermissionDenied('Not staff and not permitted')

    # Results:
    items = list(qs[0:2])
    if not items:
        # Was this because we didn't have permission or because there
        # simply wasn't a property?
        if (request.user.has_perm('pstore.view_any_object') and
                not Property.objects.filter(
                    object__identifier=object_identifier,
                    name=property_name).exists()):
            raise Http404('No such property')
        raise PermissionDenied('Not staff or not permitted')

    if len(items) > 1:
        class_ = Property.MultipleObjectsReturned
        raise class_('get() returned more than one Property. Lookup parameters'
                     ' were %r' % ({'object': object_identifier,
                                    'name': property_name},))

    # TODO: if the value is really large, we should somehow write it to disk
    # before passing it around.
    property = items[0]
    file = BytesIO(property.value)

    # Response:
    return EncryptedResponse(fp=file, enctype=property.enctype())


@nonce_required
@require_POST
@audit_view('sets or replaces a property', mutates=True)
def set_property(request, object_identifier, property_name):
    author = request.user  # FIXME

    # Check the rest of the arguments.
    if len(request.GET) != 0:
        raise NotImplementedError('Unexpected GET args', request.GET)
    if len(request.POST) != 1:  # nonce was here..
        raise NotImplementedError('Unexpected POST args', request.POST)

    # Docode URI arguments.
    object_identifier = urlunquote(object_identifier)
    property_name = urlunquote(property_name)

    # Which users?
    usernames = [i.name for i in request.FILES.getlist(property_name)]
    assert usernames

    # Are they the-public-user?
    if usernames == ['*']:
        is_public = True
        users = None
        users_dict = {'*': None}  # the public-user
    elif all(i != '*' for i in usernames):
        is_public = False
        users = list(User.objects.filter(username__in=usernames))
        if len(users) != len(usernames):
            raise Exception('FIXME-EXCEPTION')
        users_dict = dict((i.username, i) for i in users)
    else:
        raise Exception('FIXME-EXCEPTION')

    # Autocreate object if it didn't exist yet. Next, make *sure* that 'author'
    # that we found in the nonce is a valid writer.
    if users:
        obj, created = Object.objects.get_or_create(
            identifier=object_identifier)
    else:
        try:
            obj = Object.objects.get(identifier=object_identifier)
        except Object.DoesNotExist:
            raise Http404('Refusing to create an object automatically')
        else:
            created = False

    if created:
        # We need a list of users if this is a new entry! Yes.. that means
        # that we *must* have an encrypted property for the machine to be
        # created. Let that be the case for now.
        assert users

        # Author must be in the list of users too.
        if author not in users:
            raise PermissionDenied('Author is not in list of users',
                                   author.username)

        # Give everyone admin perms for now. We shall devise some kind of
        # scheme to tell admins from readers apart in the future.
        for user in users:
            ObjectPerm.objects.create(object=obj, user=user, can_write=True)

    else:
        # Check that author has write powers.
        # TODO: we could allow SU-users to do stuff..
        if not obj.allowed.filter(can_write=True, user=author).exists():
            raise PermissionDenied('Author has no write powers for object',
                                   author.username)

        if not is_public:
            # The amount of received files must obviously be equal to the
            # number of allowed users. Checks against the individual files is
            # done below. We compare the user list to make sure we don't get
            # too few or too many.
            if set(i.user for i in obj.allowed.all()) != set(users):
                raise PermissionDenied('Object exists already and supplied '
                                       'list of users differs')

    # TODO: extra audit stuff here. do more?
    oldies = list(Property.objects.filter(object=obj, name=property_name)[0:1])
    if oldies:
        if oldies[0].type == Property.TYPE_PUBLIC and not is_public:
            # LOG: switching property from public to shared
            pass
        elif oldies[0].type == Property.TYPE_SHARED and is_public:
            # LOG: switching property from shared to public
            pass

    # Purge the old properties..
    Property.objects.filter(object=obj, name=property_name).delete()
    # ... and create new ones.
    for file in request.FILES.getlist(property_name):
        user = users_dict[file.name]
        create_property(object=obj, property=property_name, file=file,
                        user=user)

    return VoidResponse()


@nonce_required
@require_POST
@audit_view('updates properties and/or permissions', mutates=True)
def update_properties(request, object_identifier):
    if len(request.GET) != 0:
        raise NotImplementedError('Unexpected GET args', request.GET)
    if len(request.POST) != 1:  # only the nonce_b64 should be here
        raise NotImplementedError('Unexpected POST args', request.POST)

    # Decode object_identifier:
    object_identifier = urlunquote(object_identifier)

    # Check authorization and existence:
    if request.user.has_perm('pstore.view_any_object'):
        obj = get_object_or_404(Object, identifier=object_identifier)
    else:
        obj = get_object_or_403(Object, identifier=object_identifier)
    try:
        ObjectPerm.objects.get(object=obj, user=request.user, can_write=True)
    except ObjectPerm.DoesNotExist:
        raise PermissionDenied('No permissions for user', request.user)

    # Fetch users and do a few basic checks on them. Below, during the Property
    # create loop, we'll do some more checks.
    usernames = []
    for property_name in request.FILES.keys():
        for file in request.FILES.getlist(property_name):
            username = file.name
            usernames.append(username)
    usernames = set(usernames)
    users = dict([(i.username, i) for i in
                  User.objects.filter(username__in=usernames)])
    if len(users) != len(usernames):
        raise Exception('FIXME-EXCEPTION: user count invalid')

    # XXX: assert that user is in usernames!

    # TODO: do more assertions on the users here..
    # TODO: add to audit log the attempted move of userset X to userset Y

    # Fetch files and re-set properties. Here we won't accept transparent
    # switches of public to shared properties.
    for property_name in request.FILES.keys():
        # TODO: ensure that we have a file for every user for every key
        # TODO: check public/shared properties for swapperony
        Property.objects.filter(object=obj, name=property_name).delete()

        for file in request.FILES.getlist(property_name):
            # We only add new shared (encrypted) properties, we don't need to
            # touch the public ones.
            user = users[file.name]
            create_property(object=obj, property=property_name, file=file,
                            user=user)

    # Update permissions. For now, we'll only use the can_write, since we don't
    # have a means of communicating that state yet.
    # TODO: (a) audit log here?
    # TODO: (b) don't delete permitted users and add a created column to
    # ObjectPerm?
    # TODO: (c) double check that there are no user-properties for disallowed
    # users left?
    currently_allowed = set([i.user for i in
                             (ObjectPerm.objects.filter(object=obj)
                              .select_related('user'))])
    del currently_allowed  # fixme.. use this
    # XXX: take currently_allowed, remove now-allowed:
    # delete the leftovers
    # add the not-in-currently_allowed
    ObjectPerm.objects.filter(object=obj).delete()
    for user in users.values():
        ObjectPerm.objects.create(object=obj, user=user, can_write=True)

    return VoidResponse()



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2010,2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
import time

from django.conf import settings
from django.db import connection
from django.http import HttpResponse

from pstore.http import HttpError
from pstore.security import validate_nonce_b64


class AuthenticateByNonceMiddleware(object):
    def process_request(self, request):
        if request.method == 'GET':
            nonce_b64 = request.GET.get('nonce_b64')
        elif request.method == 'POST':
            nonce_b64 = request.POST.get('nonce_b64')
        else:
            nonce_b64 = None

        if nonce_b64:
            assert not hasattr(request, '_cached_user')
            request.user = validate_nonce_b64(nonce_b64)
            request.user.used_nonce = True
        else:
            request.user.used_nonce = False

        return None


class HttpErrorMiddleware(object):
    def process_exception(self, request, exception):
        if isinstance(exception, HttpError):
            return HttpResponse(content=exception.user_description,
                                content_type='text/plain; charset=utf-8',
                                status=exception.status_code)
        return None


class LogSqlToConsoleMiddleware(object):
    """
    Log all SQL statements direct to the console (in debug mode only).
    Intended for use with the django development server.

    Insert as first element in MIDDLEWARE_CLASSES when you need it.

    http://www.djangosnippets.org/snippets/1672/ by davepeck, 6-aug-2009.
    """
    def process_request(self, request):
        self.t0 = time.time()
        return None

    def process_response(self, request, response):
        if (settings.DEBUG and
                connection.queries and
                (not settings.MEDIA_URL or
                 not request.META['PATH_INFO'].startswith(
                     settings.MEDIA_URL)) and
                not request.META['PATH_INFO'].startswith('/jsi18n/')):
            print '\n' + '=' * 72

            if 'time' in connection.queries[0]:
                total = sum(float(q['time']) for q in connection.queries)
                for i, query in enumerate(connection.queries):
                    print '>>> (%d) %ss: %s' % (i, query['time'], query['sql'])
                print ('== %d queries in %f seconds ==\n' %
                       (len(connection.queries), total))

            else:
                for i, query in enumerate(connection.queries):
                    print '>>> (%d): %s' % (i, query['sql'])
                print ('== %d queries in %f seconds ==' %
                       (len(connection.queries), time.time() - self.t0))

        return response



# vim: set ts=8 sw=4 sts=4 et ai tw=79 syn=python:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.

Example WSGI config for django-pstore project.

This module contains the WSGI application used by Django's development server
and any production WSGI deployments. It should expose a module-level variable
named ``application``. Django's ``runserver`` and ``runfcgi`` commands discover
this application via the ``WSGI_APPLICATION`` setting.

Usually you will have the standard Django WSGI application here, but it also
might make sense to replace the whole Django WSGI application with a custom one
that later delegates to the Django one. For example, you could introduce WSGI
middleware here, or combine a Django application with an application of another
framework.

"""
import os
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'pstore.settings')

###############################################################################
# If you're using virtualenv(1) to manage your python dependencies, you
# can set the sys.path from your project virtualenv here:
# #import sys
# #sys.path[:] = [/* different path list */]
# But your wsgi server might be in the know about virtualenv. E.g. for uwsgi
# you can do this in the config:
# #virtualenv = /srv/virtualenvs/pstore
# Possibly combined with the PYTHONPATH environment variable:
# #env = PYTHONPATH=/srv/django-apps:/opt/django14
###############################################################################

# This application object is used by any WSGI server configured to use this
# file. This includes Django's development server, if the WSGI_APPLICATION
# setting points here.
try:
    from django.core.wsgi import get_wsgi_application
except ImportError:  # django1.3- ?
    from django.core.handlers.wsgi import WSGIHandler as get_wsgi_application
application = get_wsgi_application()



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2010,2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
from django.contrib import admin
from django.contrib.auth.models import User

from pstore.models import Object, ObjectPerm, Property, PublicKey

try:
    # Django 1.4+
    from django.contrib.admin import SimpleListFilter
except ImportError:
    # Django 1.3-
    ObjectUserFilter = None
else:
    class ObjectUserFilter(SimpleListFilter):
        title = 'user'
        parameter_name = 'user'  # used in url

        def lookups(self, request, model_admin):
            """
            Returns elements for query filter on the side. Usernames are added
            both with and without a plus, where the former one means: "select
            only records that match *only* that user".
            """
            users = []
            for user in (User.objects.order_by('-is_active', 'username')
                         .only('id', 'username')):
                users.append((user.id, user.username))
                users.append(('%d+' % (user.id,), '%s+' % (user.username,)))
            return users

        def queryset(self, request, queryset):
            value = self.value()
            if not value:
                return queryset

            # If the value ends with a '+', the user must be in the allowed
            # list.
            if value.endswith('+'):
                value = value[0:-1]
                match_only = False
            # If it doesn't, the user must be the only one in the allowed list.
            else:
                match_only = True

            queryset = queryset.filter(allowed__user__id=value)
            if match_only:
                # Filter by records that have only one allowed user.
                queryset = queryset.extra(
                    where=[('(SELECT COUNT(*) FROM pstore_objectperm '
                            'WHERE object_id = pstore_object.id) = 1')]
                )

            return queryset


class ObjectAdmin(admin.ModelAdmin):
    def users(self, object):
        all = [i.user.username + ('', '*')[i.can_write]
               for i in (ObjectPerm.objects.filter(object=object)
                         .select_related('user'))]
        return ', '.join(sorted(all))

    list_display = ('identifier', 'users')
    ordering = ('identifier',)
    search_fields = ('identifier',)

    if ObjectUserFilter:
        list_filter = (ObjectUserFilter,)


class ObjectPermAdmin(admin.ModelAdmin):
    list_display = ('object', 'user', 'can_write')


class PropertyAdmin(admin.ModelAdmin):
    def short_value(self, object):
        if (object.type & object.BIT_ENCRYPTED):
            return '(encrypted)'

        value = object.value[0:32]
        if len(value) >= 32:
            value = value + '...'
        return repr(value)  # makes sure we won't see non-ascii

    exclude = ('value',)  # raw/binary values.. should not be in admin
    list_display = ('id', 'object', 'name', 'short_value', 'user')
    ordering = ('object__identifier', 'name', 'user__username')
    search_fields = ('object__identifier', 'name')


class PublicKeyAdmin(admin.ModelAdmin):
    list_display = ('__unicode__', 'description')
    ordering = ('user__username',)


admin.site.register(Object, ObjectAdmin)
admin.site.register(ObjectPerm, ObjectPermAdmin)
admin.site.register(Property, PropertyAdmin)
admin.site.register(PublicKey, PublicKeyAdmin)



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
import logging

from django.core.servers.basehttp import FileWrapper
from django.http import HttpResponse

from pstorelib.bytes import get_size


logger = logging.getLogger('pstore')


class HttpError(Exception):
    """
    This exception is passed on to the HTTP layer where it is propagated back
    to the client.
    """
    def __init__(self, status_code, user_description, error_description=None):
        """
        The status_code is the HTTP error code.
        The user_description is the problem description returned over HTTP.
        Optionally error_description may hold a text that gets sent to the
        error log.
        """
        super(HttpError, self).__init__(status_code, user_description)
        self.status_code = status_code
        self.user_description = user_description

        if error_description:
            # Passing args instead of instance, as we'd otherwise lose the
            # error_description in the error mail.
            args = (status_code, user_description, error_description)
            sys_exc_info = (HttpError, args, None)
            logger.error(error_description, exc_info=sys_exc_info)


class EncryptedResponse(HttpResponse):
    """
    An HTTP response returning encrypted data. It doesn't do any encryption, it
    just adds the appropriate headers for the decrypting end to know how to
    decrypt it.
    """
    def __init__(self, data=None, fp=None, enctype=None):
        """
        Specify either file or data.
        """
        assert bool(data) ^ bool(fp)
        assert enctype in ('none', 'gpg', 'sshrsa')

        if data:
            assert isinstance(data, str)  # .. and not unicode
            content = data
            content_length = len(data)
        else:
            content = FileWrapper(fp)
            content_length = get_size(fp)
            assert content_length != -1

        ctype = 'application/octet-stream'
        super(EncryptedResponse, self).__init__(content, content_type=ctype)
        self['Content-Length'] = content_length
        self['X-Encryption'] = enctype


class VoidResponse(HttpResponse):
    """
    An HTTP response without any content.
    """
    def __init__(self):
        content = 'OK'
        ctype = 'text/plain; charset=us-ascii'

        # Status: 204 No Content
        super(VoidResponse, self).__init__(content_type=ctype,
                                           status=204)
        self['Content-Length'] = len(content)
